#pragma once

#include "LuaStackChecker.h"
#include "LuaStackGuard.h"
#include "lua/lua.hpp"
#include <map>
#include <set>
#include <string>
#include <vector>

template<typename T>
void toLua(lua_State* L, const T& arg); //  .      .

void toLua(lua_State* L, void* ptr);

template<typename T>
void toLua(lua_State* L, const std::vector<T>& arg)
{
  // stack:
  lua_newtable(L); // stack: table
  const size_t size = arg.size();
  for(size_t i = 0; arg.size() > i; ++i)
  {
    lua_pushnumber(L, i + 1); // stack: table i
    toLua(L, arg[i]); // stack: table i value
    lua_settable(L, -3); // stack: table
  }
}

template<typename T>
void toLua(lua_State* L, const std::set<T>& arg)
{
  // stack:
  lua_newtable(L); // stack: table
  int counter = 1;
  for(typename std::set<T>::const_iterator i = arg.begin(); arg.end() != i; ++i, ++counter)
  {
    lua_pushnumber(L, counter); // stack: table counter
    toLua(L, *i); // stack: table counter value
    lua_settable(L, -3); // stack: table
  }
}

template<typename TKey, typename TValue>
void toLua(lua_State* L, const std::map<TKey, TValue>& arg)
{
  // stack:
  lua_newtable(L); // stack: table
  for(typename std::map<TKey, TValue>::const_iterator i = arg.begin(); arg.end() != i; ++i)
  {
    toLua(L, i->first); // stack: table key
    toLua(L, i->second); // stack: table key value
    lua_settable(L, -3); // stack: table
  }
}

template<typename T>
bool fromLua(lua_State* L, int index, T& ret); //  .      .

template<typename T>
bool fromLua(lua_State* L, int index, std::vector<T>& ret)
{
  if(!lua_istable(L, index))
    return false;

  LuaStackChecker sc(L, __FILE__, __LINE__);

  lua_pushvalue(L, index); // stack: vector

  const int count = luaL_getn(L, -1);
  for(int i = 1; count >= i; ++i)
  {
    lua_pushnumber(L, i);
    lua_gettable(L, -2);
    T value;
    fromLua(L, -1, value);
    ret.push_back(value);
    lua_pop(L, 1); // stack: vector
  }
  lua_pop(L, 1); // stack:

  return true;
}

template<typename T>
bool fromLua(lua_State* L, int index, std::set<T>& ret)
{
  if(!lua_istable(L, index))
    return false;

  LuaStackChecker sc(L, __FILE__, __LINE__);

  lua_pushvalue(L, index); // stack: set

  const int count = luaL_getn(L, -1);

  for(int i = 1; count >= i; ++i)
  {
    lua_pushnumber(L, i);
    lua_gettable(L, -2);
    T value;
    fromLua(L, -1, value);
    ret.insert(value);
    lua_pop(L, 1);
  }
  lua_pop(L, 1); // stack:

  return true;
}

template<typename TKey, typename TValue>
bool fromLua(lua_State* L, int index, std::map<TKey, TValue>& ret)
{
  // stack:
  if(!lua_istable(L, index))
    return false;

  LuaStackChecker sc(L, __FILE__, __LINE__);

  lua_pushvalue(L, index); // stack: map

  lua_pushnil(L); // stack: map nil
  while(lua_next(L, -2)) // stack: map key value
  {
    TKey key;
    fromLua(L, -2, key);
    TValue value;
    fromLua(L, -1, value);
    ret[key] = value;
    lua_pop(L, 1); // stack: map key
  }
  // stack: map
  lua_pop(L, 1); // stack:

  return true;
}

template<typename T>
bool fromLua(lua_State* L, int index, const char* key, T& ret)
{
  // stack: table
  LuaStackChecker sc(L, __FILE__, __LINE__);
  lua_getfield(L, index? index : LUA_GLOBALSINDEX, key); // stack: table value
  bool res = fromLua(L, -1, ret);
  lua_pop(L, 1); // stack: table

  return res;
}

// 0 args 0 returns
bool luaT(lua_State* L, const char* function);

// 0 args 1 returns
template<typename TRet>
bool luaT(lua_State* L, const char* function, TRet& ret)
{
  LuaStackChecker sc(L, __FILE__, __LINE__);
                                         // stack:
  lua_getglobal(L, function); // stack: function
  if(lua_pcall(L, 0, 1, 0)) // stack: ret
  {
    const char* err = lua_tostring(L, -1);
    lua_pop(L, 1);
    return false;
  }
  fromLua(L, -1, ret);
  lua_pop(L, 1);
  return true;
}

// 0 args 2 returns
template<typename TRet1, typename TRet2>
bool luaT(lua_State* L, const char* function, TRet1& ret1, TRet2& ret2)
{
  LuaStackChecker sc(L, __FILE__, __LINE__);
                                         // stack:
  lua_getglobal(L, function); // stack: function
  if(lua_pcall(L, 0, 2, 0)) // stack: ret1 ret2
  {
    const char* err = lua_tostring(L, -1);
    lua_pop(L, 1);
    return false;
  }
  fromLua(L, -2, ret1);
  fromLua(L, -1, ret2);
  lua_pop(L, 2);
  return true;
}

// 0 args 3 returns
template<typename TRet1, typename TRet2, typename TRet3>
bool luaT(lua_State* L, const char* function, TRet1& ret1, TRet2& ret2, TRet3& ret3)
{
  LuaStackChecker sc(L, __FILE__, __LINE__);
                                         // stack:
  lua_getglobal(L, function); // stack: function
  if(lua_pcall(L, 0, 3, 0)) // stack: ret1 ret2
  {
    const char* err = lua_tostring(L, -1);
    lua_pop(L, 1);
    return false;
  }
  fromLua(L, -3, ret1);
  fromLua(L, -2, ret2);
  fromLua(L, -1, ret3);
  lua_pop(L, 3);
  return true;
}

// 1 args 0 returns
template<typename TArg>
bool luaT(lua_State* L, const char* function, const TArg& arg)
{
  LuaStackChecker sc(L, __FILE__, __LINE__);
                                         // stack:
  lua_getglobal(L, function); // stack: function
  toLua(L, arg); // stack: function arg
  if(lua_pcall(L, 1, 0, 0)) // stack:
  {
    const char* err = lua_tostring(L, -1);
    lua_pop(L, 1);
    return false;
  }
  return true;
}

// 1 args 1 returns
template<typename TArg, typename TRet>
bool luaT(lua_State* L, const char* function, const TArg& arg, TRet& ret)
{
  LuaStackChecker sc(L, __FILE__, __LINE__);
                                         // stack:
  lua_getglobal(L, function); // stack: function
  toLua(L, arg); // stack: function arg
  if(lua_pcall(L, 1, 1, 0)) // stack: ret
  {
    const char* err = lua_tostring(L, -1);
    lua_pop(L, 1);
    return false;
  }
  fromLua(L, -1, ret);
  lua_pop(L, 1);
  return true;
}

// 1 args 2 returns
template<typename TArg, typename TRet1, typename TRet2>
bool luaT(lua_State* L, const char* function, const TArg& arg, TRet1& ret1, TRet2& ret2)
{
  LuaStackChecker sc(L, __FILE__, __LINE__);
                                         // stack:
  lua_getglobal(L, function); // stack: function
  toLua(L, arg); // stack: function arg
  if(lua_pcall(L, 1, 2, 0)) // stack: ret
  {
    const char* err = lua_tostring(L, -1);
    lua_pop(L, 1);
    return false;
  }
  fromLua(L, -1, ret1);
  fromLua(L, -2, ret2);
  lua_pop(L, 2);
  return true;
}

// 2 args 0 returns
template<typename TArg1, typename TArg2>
bool luaT(lua_State* L, const char* function, const TArg1& arg1, const TArg2& arg2)
{
  LuaStackChecker sc(L, __FILE__, __LINE__);
                                         // stack:
  lua_getglobal(L, function); // stack: function
  toLua(L, arg1); // stack: function arg1
  toLua(L, arg2); // stack: function arg1 arg2
  if(lua_pcall(L, 2, 0, 0)) // stack:
  {
    const char* err = lua_tostring(L, -1);
    lua_pop(L, 1);
    return false;
  }
  return true;
}

// 2 args 1 returns
template<typename TArg1, typename TArg2, typename TRet>
bool luaT(lua_State* L, const char* function, const TArg1& arg1, const TArg2& arg2, TRet& ret)
{
  LuaStackChecker sc(L, __FILE__, __LINE__);
                                         // stack:
  lua_getglobal(L, function); // stack: function
  toLua(L, arg1); // stack: function arg1
  toLua(L, arg2); // stack: function arg1 arg2
  if(lua_pcall(L, 2, 1, 0)) // stack: ret
  {
    const char* err = lua_tostring(L, -1);
    lua_pop(L, 1);
    return false;
  }
  fromLua(L, -1, ret);
  lua_pop(L, 1);
  return true;
}

// 3 args 0 returns
template<typename TArg1, typename TArg2, typename TArg3>
void luaT(lua_State* L, const char* function, const TArg1& arg1, const TArg2& arg2, const TArg3& arg3)
{
  LuaStackChecker sc(L, __FILE__, __LINE__);
                                         // stack:
  lua_getglobal(L, function); // stack: function
  toLua(L, arg1); // stack: function arg1
  toLua(L, arg2); // stack: function arg1 arg2
  toLua(L, arg3); // stack: function arg1 arg2 arg3
  if(lua_pcall(L, 3, 0, 0)) // stack:
  {
    const char* err = lua_tostring(L, -1);
    lua_pop(L, 1);
    return false;
  }
  return true;
}

// 3 args 1 returns
template<typename TArg1, typename TArg2, typename TArg3, typename TRet>
bool luaT(lua_State* L, const char* function, const TArg1& arg1, const TArg2& arg2, const TArg3& arg3, TRet& ret)
{
  LuaStackChecker sc(L, __FILE__, __LINE__);
                                         // stack:
  lua_getglobal(L, function); // stack: function
  toLua(L, arg1); // stack: function arg1
  toLua(L, arg2); // stack: function arg1 arg2
  toLua(L, arg3); // stack: function arg1 arg2 arg3
  if(lua_pcall(L, 3, 1, 0)) // stack: ret
  {
    const char* err = lua_tostring(L, -1);
    lua_pop(L, 1);
    return false;
  }
  fromLua(L, -1, ret);
  lua_pop(L, 1);
  return true;
}

// 4 args 0 returns
template<typename TArg1, typename TArg2, typename TArg3, typename TArg4>
bool luaT(lua_State* L, const char* function, const TArg1& arg1, const TArg2& arg2, const TArg3& arg3, const TArg4& arg4)
{
  LuaStackChecker sc(L, __FILE__, __LINE__);
                                         // stack:
  lua_getglobal(L, function); // stack: function
  toLua(L, arg1); // stack: function arg1
  toLua(L, arg2); // stack: function arg1 arg2
  toLua(L, arg3); // stack: function arg1 arg2 arg3
  toLua(L, arg4); // stack: function arg1 arg2 arg3 arg4
  if(lua_pcall(L, 4, 0, 0)) // stack:
  {
    const char* err = lua_tostring(L, -1);
    lua_pop(L, 1);
    return false;
  }
  return true;
}

// 4 args 1 returns
template<typename TArg1, typename TArg2, typename TArg3, typename TArg4, typename TRet>
bool luaT(lua_State* L, const char* function, const TArg1& arg1, const TArg2& arg2, const TArg3& arg3, const TArg4& arg4, TRet& ret)
{
  LuaStackChecker sc(L, __FILE__, __LINE__);
                                         // stack:
  lua_getglobal(L, function); // stack: function
  toLua(L, arg1); // stack: function arg1
  toLua(L, arg2); // stack: function arg1 arg2
  toLua(L, arg3); // stack: function arg1 arg2 arg3
  toLua(L, arg4); // stack: function arg1 arg2 arg3 arg4
  if(lua_pcall(L, 4, 1, 0)) // stack: ret
  {
    const char* err = lua_tostring(L, -1);
    lua_pop(L, 1);
    return false;
  }
  fromLua(L, -1, ret);
  lua_pop(L, 1);
  return true;
}

