#include "LuaUtil.h"
#include "glew.h"
#include "SOIL.h"
#include <vector>
#include <string>

using namespace std;

// stack: opengl
#define glConst(L, name) lua_pushnumber(L, GL_##name);\
                         lua_setfield(L, -2, #name)\

int getErrors(lua_State* L)
{
  GLenum err = glGetError();
  vector<string> errors;

  while(GL_NO_ERROR != err)
  {
    errors.push_back((const char*)gluErrorString(err));
  }

  if(errors.empty())
  {
    return 0;
  }

  toLua(L, errors);

  return 1;
}

void registerConstants(lua_State* L)
{
  // stack:
  LuaStackChecker sc(L, __FILE__, __LINE__);
  lua_getglobal(L, "opengl"); // stack: opengl
  #include "glconst.inl"
  lua_pop(L, 1); // stack:
}

int create(lua_State* L)
{
  // stack:
  glewInit();
  registerConstants(L);

  return 0;
}

int clear(lua_State* L)
{
  // stack:
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  return 0;
}

int matrixMode(lua_State* L)
{
  // stack: mode
  int mode;
  if(fromLua(L, 1, mode))
  {
    glMatrixMode(mode);
  }
  else
  {
    luaL_error(L, "Invalid arguments!");
  }

  return 0;
}

int pushMatrix(lua_State* L)
{
  // stack:
  glPushMatrix();

  return 0;
}

int popMatrix(lua_State* L)
{
  glPopMatrix();

  return 0;
}

int enable(lua_State* L)
{
  // state: param
  int param;
  if(fromLua(L, 1, param))
  {
    glEnable(param);
  }
  else
  {
    luaL_error(L, "Invalid arguments!");
  }

  return 0;
}

int disable(lua_State* L)
{
  // state: param
  int param;
  if(fromLua(L, 1, param))
  {
    glDisable(param);
  }
  else
  {
    luaL_error(L, "Invalid arguments!");
  }

  return 0;
}

int viewport(lua_State* L)
{
  // stack: x, y, w, h
  int x, y, w, h;
  if(fromLua(L, 1, x) &&
     fromLua(L, 2, y) &&
     fromLua(L, 3, w) &&
     fromLua(L, 4, h))
  {
    glViewport(x, y, w, h);
  }
  else
  {
    luaL_error(L, "Invalid arguments!");
  }

  return 0;
}

int loadIdentity(lua_State* L)
{
  // stack:
  glLoadIdentity();

  return 0;
}

int uPerspective(lua_State* L)
{
  // stack: fov, aspect, near, far
  // near  far    
  // (    MSVS)
  double fov, aspect, n, f;
  if(fromLua(L, 1, fov) &&
     fromLua(L, 2, aspect) &&
     fromLua(L, 3, n) &&
     fromLua(L, 4, f))
  {
    gluPerspective(fov, aspect, n, f);
  }
  else
  {
    luaL_error(L, "Invalid arguments!");
  }

  return 0;
}

int translate(lua_State* L)
{
  // stack: x, y, z
  double x, y, z;
  if(fromLua(L, 1, x) &&
     fromLua(L, 2, y) &&
     fromLua(L, 3, z))
  {
    glTranslated(x, y, z);
  }
  else
  {
    luaL_error(L, "Invalid arguments!");
  }

  return 0;
}

int rotate(lua_State* L)
{
  // stack: angle, x, y, z
  double angle, x, y, z;
  if(fromLua(L, 1, angle) &&
     fromLua(L, 2, x) &&
     fromLua(L, 3, y) &&
     fromLua(L, 4, z))
  {
    glRotated(angle, x, y, z);
  }
  else
  {
    luaL_error(L, "Invalid arguments!");
  }

  return 0;
}

int begin(lua_State* L)
{
  // state: mode
  int mode;
  if(fromLua(L, 1, mode))
  {
    glBegin(mode);
  }
  else
  {
    luaL_error(L, "Invalid arguments!");
  }

  return 0;
}

int end(lua_State* L)
{
  // stack:
  glEnd();

  return 0;
}

int shadeModel(lua_State* L)
{
  // state: model
  int model;
  if(fromLua(L, 1, model))
  {
    glShadeModel(model);
  }
  else
  {
    luaL_error(L, "Invalid arguments!");
  }

  return 0;
}

int clearColor(lua_State* L)
{
  // stack: angle, x, y, z
  double r, g, b, a;
  if(fromLua(L, 1, r) &&
     fromLua(L, 2, g) &&
     fromLua(L, 3, b) &&
     fromLua(L, 4, a))
  {
    glClearColor(r, g, b, a);
  }
  else
  {
    luaL_error(L, "Invalid arguments!");
  }

  return 0;
}

int clearDepth(lua_State* L)
{
  // state: depth
  int depth;
  if(fromLua(L, 1, depth))
  {
    glClearDepth(depth);
  }
  else
  {
    luaL_error(L, "Invalid arguments!");
  }

  return 0;
}

int depthFunc(lua_State* L)
{
  // state: func
  int func;
  if(fromLua(L, 1, func))
  {
    glDepthFunc(func);
  }
  else
  {
    luaL_error(L, "Invalid arguments!");
  }

  return 0;
}

int hint(lua_State* L)
{
  // stack: target mode
  int target, mode;
  if(fromLua(L, 1, target) &&
     fromLua(L, 2, mode))
  {
    glHint(target, mode);
  }
  else
  {
    luaL_error(L, "Invalid arguments!");
  }

  return 0;
}

int texCoord2(lua_State* L)
{
  // stack: s t
  double s, t;
  if(fromLua(L, 1, s) &&
     fromLua(L, 2, t))
  {
    glTexCoord2d(s, t);
  }
  else
  {
    luaL_error(L, "Invalid arguments!");
  }

  return 0;
}

int vertex(lua_State* L)
{
  // stack: x, y, z
  double x, y, z = 0; // z - 
  if(fromLua(L, 1, x) &&
     fromLua(L, 2, y))
  {
    fromLua(L, 3, z);
    glVertex3d(x, y, z);
  }
  else
  {
    luaL_error(L, "Invalid arguments!");
  }

  return 0;
}

int color(lua_State* L)
{
  // stack: r, g, b, a
  double r, g, b, a = 1; // a - 
  if(fromLua(L, 1, r) &&
     fromLua(L, 2, g) &&
     fromLua(L, 3, b))
  {
    fromLua(L, 4, a);
    glColor4d(r, g, b, a);
  }
  else
  {
    luaL_error(L, "Invalid arguments!");
  }

  return 0;
}

int loadImageData(lua_State* L)
{
  // stack: filename
  string filename;
  if(fromLua(L, 1, filename))
  {
    int width, height, channels;
    unsigned char* data = SOIL_load_image(filename.c_str(), &width, &height, &channels,SOIL_LOAD_AUTO);
    if(data)
    {
      lua_pushlightuserdata(L, (void*)data); // stack: filename data
      lua_pushnumber(L, width); // stack: filename data width
      lua_pushnumber(L, height); // stack: filename data width height
      lua_pushnumber(L, channels); // stack: filename data width height channels

      return 4;
    }
    else
    {
      printf( "Image loading error: '%s'\n", SOIL_last_result() );

      return 0;
    }
  }
  else
  {
    luaL_error(L, "Invalid arguments!");
  }

  return 0;
}

void toLua(lua_State* L, const GLuint& arg)
{
  lua_pushnumber(L, arg);
}

int genTextures(lua_State* L)
{
  // stack: count
  int count;
  if(fromLua(L, 1, count))
  {
    vector<GLuint> textures(count, 0);
    glGenTextures(count, &textures[0]);
    toLua(L, textures);

    return 1;
  }
  else
  {
    luaL_error(L, "Invalid arguments!");
  }

  return 0;
}

int bindTexture(lua_State* L)
{
  // stack: target texture
  int target, texture;
  if(fromLua(L, 1, target) &&
     fromLua(L, 2, texture))
  {
    glBindTexture(target, texture);
  }
  else
  {
    luaL_error(L, "Invalid arguments!");
  }

  return 0;
}

int texImage2D(lua_State* L)
{
  // stack: target level internalFormat width height border format data
  int target, level, internalFormat, width, height, border, format;
  unsigned char* data;
  if(fromLua(L, 1, target) &&
     fromLua(L, 2, level) &&
     fromLua(L, 3, internalFormat) &&
     fromLua(L, 4, width) &&
     fromLua(L, 5, height) &&
     fromLua(L, 6, border) &&
     fromLua(L, 7, format))
  {
    data = (unsigned char*)lua_touserdata(L, 8);
    if(!data)
    {
      luaL_error(L, "Invalid arguments!");
    }
    glTexImage2D(target, level, internalFormat, width, height, border, format, GL_UNSIGNED_BYTE, data);
  }
  else
  {
    luaL_error(L, "Invalid arguments!");
  }

  return 0;
}

int texParameteri(lua_State* L)
{
  // stack: target pname param
  int target, pname, param;
  if(fromLua(L, 1, target) &&
     fromLua(L, 2, pname) &&
     fromLua(L, 3, param))
  {
    glTexParameteri(target, pname, param);
  }
  else
  {
    luaL_error(L, "Invalid arguments!");
  }

  return 0;
}

const luaL_Reg opengl[] = {
  {"Create", create},
  {"GetErrors", getErrors},
  {"Clear", clear},
  {"Enable", enable},
  {"Disable", disable},
  {"Viewport", viewport},
  {"uPerspective", uPerspective},
  {"MatrixMode", matrixMode},
  {"PushMatrix", pushMatrix},
  {"PopMatrix", popMatrix},
  {"LoadIdentity", loadIdentity},
  {"Translate", translate},
  {"Rotate", rotate},
  {"Begin", begin},
  {"End", end},
  {"ShadeModel", shadeModel},
  {"ClearColor", clearColor},
  {"ClearDepth", clearDepth},
  {"DepthFunc", depthFunc},
  {"Hint", hint},
  {"TexCoord2", texCoord2},
  {"Vertex", vertex},
  {"Color", color},
  {"LoadImageData", loadImageData},
  {"GenTextures", genTextures},
  {"BindTexture", bindTexture},
  {"TexImage2D", texImage2D},
  {"TexParameteri", texParameteri},
  {0, 0},
};

extern "C" __declspec(dllexport) int luaopen_opengl(lua_State* L)
{
	luaL_register(L, "opengl", opengl);
  return 1;
}
